package cn.com.ty.chat.neety.server;

import cn.com.ty.chat.utils.RemotingUtil;
import io.netty.bootstrap.ServerBootstrap;
import io.netty.channel.*;
import io.netty.channel.epoll.Epoll;
import io.netty.channel.epoll.EpollEventLoopGroup;
import io.netty.channel.epoll.EpollServerSocketChannel;
import io.netty.channel.nio.NioEventLoopGroup;
import io.netty.channel.socket.nio.NioServerSocketChannel;
import io.netty.handler.ssl.SslContext;
import io.netty.handler.ssl.SslContextBuilder;
import lombok.extern.slf4j.Slf4j;

import javax.net.ssl.KeyManagerFactory;
import javax.net.ssl.SSLException;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.security.KeyStore;
import java.security.KeyStoreException;
import java.security.NoSuchAlgorithmException;
import java.security.UnrecoverableKeyException;
import java.security.cert.CertificateException;

@Slf4j
public class ChatServer {

    // boss线程
    private EventLoopGroup bossGroup = null;
    // worker线程
    private EventLoopGroup workerGroup = null;

    private boolean isLinuxPlatform() {
        return RemotingUtil.isLinuxPlatform() && Epoll.isAvailable();
    }

    /**
     * 启动Netty Server
     *
     * @throws InterruptedException
     */
    public void start(Integer port, Integer readerIdleTime, String sslPath, String sslPassword, Boolean sslEnable) throws InterruptedException {
        try {
            if (isLinuxPlatform()) {
                bossGroup = new EpollEventLoopGroup(1);
                workerGroup = new EpollEventLoopGroup(100);
            } else {
                bossGroup = new NioEventLoopGroup(1);
                workerGroup = new NioEventLoopGroup(100);
            }
            SslContext sslContext = null;

            if (sslEnable) {
                //构建ssl，适配wss请求
                KeyStore keyStore = KeyStore.getInstance("JKS");
                keyStore.load(new FileInputStream(sslPath), sslPassword.toCharArray());

                KeyManagerFactory keyManagerFactory = KeyManagerFactory.getInstance("SunX509");
                keyManagerFactory.init(keyStore, sslPassword.toCharArray());

                sslContext = SslContextBuilder.forServer(keyManagerFactory).build();
            }
            // netty服务,启动引擎
            ServerBootstrap b = new ServerBootstrap();
            // 主从模式
            b.group(bossGroup, workerGroup)
                // 主线程处理类
                .channel(isLinuxPlatform() ? EpollServerSocketChannel.class : NioServerSocketChannel.class)
                // 配置信息
                .option(ChannelOption.SO_BACKLOG, 1024)// 针对主线程配置
                // .childOption(ChannelOption.SO_KEEPALIVE,true);//子线程配置

                // 子线程的处理，Handler
                .childHandler(new MyInitChannel(sslContext, readerIdleTime));
            // 等待客户端连接
            ChannelFuture future = b.bind(port).sync();
            log.info("websocket-server started successful. port:{} ", port);
            future.channel().closeFuture().sync();
        } catch (FileNotFoundException e) {
            e.printStackTrace();
        } catch (KeyStoreException e) {
            e.printStackTrace();
        } catch (NoSuchAlgorithmException e) {
            e.printStackTrace();
        } catch (CertificateException e) {
            e.printStackTrace();
        } catch (SSLException e) {
            e.printStackTrace();
        } catch (UnrecoverableKeyException e) {
            e.printStackTrace();
        } catch (IOException e) {
            e.printStackTrace();
        } finally {
            bossGroup.shutdownGracefully().sync();
            workerGroup.shutdownGracefully().sync();
        }
    }
}
