package cn.com.ty.chat.neety.server.handler;

import cn.com.ty.chat.neety.process.ProcessorFactory;
import cn.com.ty.chat.neety.process.ProcessorInterface;
import cn.com.ty.chat.neety.protocol.IMDecoder;
import cn.com.ty.chat.neety.protocol.IMMessage;
import cn.com.ty.chat.neety.support.WebSocketSupport;
import io.netty.buffer.ByteBuf;
import io.netty.buffer.Unpooled;
import io.netty.channel.*;
import io.netty.handler.codec.http.*;
import io.netty.handler.codec.http.websocketx.*;
import io.netty.handler.ssl.SslHandshakeCompletionEvent;
import io.netty.handler.timeout.IdleState;
import io.netty.handler.timeout.IdleStateEvent;
import io.netty.util.AttributeKey;
import io.netty.util.CharsetUtil;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.lang3.StringUtils;

@Slf4j
@ChannelHandler.Sharable
public class WebSocketHandler extends SimpleChannelInboundHandler<Object> {

    private WebSocketServerHandshaker handShaker;
    private ChannelHandlerContext ctx;
    private WebSocketSupport process = new WebSocketSupport();

    @Override
    protected void channelRead0(ChannelHandlerContext ctx, Object msg) throws Exception {
        // 传统的HTTP接入
        if (msg instanceof FullHttpRequest) {
            handleHttpRequest(ctx, (FullHttpRequest)msg);
        }
        // WebSocket接入
        else if (msg instanceof WebSocketFrame) {
            handleWebSocket(ctx, (WebSocketFrame)msg);
        }
    }

    /**
     * 处理Http请求，完成WebSocket握手 注意：WebSocket连接第一次请求使用的是Http
     * 
     * @param ctx
     * @param request
     * @throws Exception
     */
    private void handleHttpRequest(ChannelHandlerContext ctx, FullHttpRequest request) throws Exception {
        // 如果HTTP解码失败，返回HHTP异常
        if (!request.decoderResult().isSuccess() || (!"websocket".equals(request.headers().get("Upgrade")))) {
            sendHttpResponse(ctx, request, new DefaultFullHttpResponse(HttpVersion.HTTP_1_1, HttpResponseStatus.BAD_REQUEST));
            return;
        }

        // 正常WebSocket的Http连接请求，构造握手响应返回
        WebSocketServerHandshakerFactory wsFactory = new WebSocketServerHandshakerFactory("wss://" + request.headers().get(HttpHeaders.Names.HOST), null, false);
        handShaker = wsFactory.newHandshaker(request);
        if (handShaker == null) { // 无法处理的websocket版本
            WebSocketServerHandshakerFactory.sendUnsupportedWebSocketVersionResponse(ctx.channel());
        } else { // 向客户端发送websocket握手,完成握手
            handShaker.handshake(ctx.channel(), request);
            // 记录管道处理上下文，便于服务器推送数据到客户端
            this.ctx = ctx;
        }
    }

    /**
     * Http返回
     * 
     * @param ctx
     * @param request
     * @param response
     */
    private static void sendHttpResponse(ChannelHandlerContext ctx, FullHttpRequest request, FullHttpResponse response) {
        // 返回应答给客户端
        if (response.getStatus().code() != 200) {
            ByteBuf buf = Unpooled.copiedBuffer(response.getStatus().toString(), CharsetUtil.UTF_8);
            response.content().writeBytes(buf);
            buf.release();
            HttpHeaders.setContentLength(response, response.content().readableBytes());
        }

        // 如果是非Keep-Alive，关闭连接
        ChannelFuture f = ctx.channel().writeAndFlush(response);
        if (!HttpHeaders.isKeepAlive(request) || response.getStatus().code() != 200) {
            f.addListener(ChannelFutureListener.CLOSE);
        }
    }

    /**
     * 处理Socket请求
     * 
     * @param ctx
     * @param frame
     * @throws Exception
     */
    private void handleWebSocket(ChannelHandlerContext ctx, WebSocketFrame frame) throws Exception {
        // 判断是否是关闭链路的指令
        if (frame instanceof CloseWebSocketFrame) {
            log.info("我在-关闭链路的指令");
            process.logout(ctx.channel());

            handShaker.close(ctx.channel(), (CloseWebSocketFrame)frame.retain());
            return;
        }
        // 判断是否是Ping消息
        if (frame instanceof PingWebSocketFrame) {
            log.info("我在-Ping消息");
            ctx.channel().write(new PongWebSocketFrame(frame.content().retain()));
            return;
        }
        // 当前只支持文本消息，不支持二进制消息
        if (!(frame instanceof TextWebSocketFrame)) {
            log.info("当前只支持文本消息，不支持二进制消息");
            throw new UnsupportedOperationException("当前只支持文本消息，不支持二进制消息");
        }
        String msg = ((TextWebSocketFrame)frame).text();

        // 处理来自客户端的WebSocket请求
        // 将字符串解析为自定义格式
        IMMessage message = IMDecoder.decode(msg);
        if (null == message) {
            return;
        }

        if (message.getUserId().equals(-1)){
            log.info("游客关闭连接，结束下面的逻辑");
            process.logout(ctx.channel());
            handShaker.close(ctx.channel(), (CloseWebSocketFrame)frame.retain());
            return;
        }
        String cmd = message.getCmd();
        String username = message.getName();
        ProcessorInterface processor = ProcessorFactory.getOperation(cmd).orElseThrow(() -> {
            String str = String.format("%s>>>发送WebSoket协议异常,%s",username,msg);
            return new RuntimeException(str);
        });
        log.info(message.toString());
        processor.process(ctx.channel(), message);
    }

    @Override
    public void channelActive(ChannelHandlerContext ctx) throws Exception {
        log.info("{}--------->已经建立连接", process.getAddress(ctx.channel()));

        super.channelActive(ctx);
    }

    @Override
    public void userEventTriggered(ChannelHandlerContext ctx, Object evt) throws Exception {
//        super.userEventTriggered(ctx, evt);
        if (evt instanceof IdleStateEvent){
            IdleStateEvent event = (IdleStateEvent)evt;
            if (event.state()== IdleState.READER_IDLE){
                log.info("{}-----》客户端操作空闲",  process.getAddress(ctx.channel()));
                process.logout(ctx.channel());
//                CloseWebSocketFrame fram = new CloseWebSocketFrame();
//                handShaker.close(ctx.channel(), fram.retain());
            }
        }
        if (evt instanceof SslHandshakeCompletionEvent){
            log.info("{}-----》ssl链接超时", process.getAddress(ctx.channel()));
            process.logout(ctx.channel());
            /*SslHandshakeCompletionEvent handshake = (SslHandshakeCompletionEvent) evt;
            if (handshake.isSuccess()) {
                ctx.fireChannelActive();
            } else{
                process.logout(ctx.channel());
            }*/
        }
    }

    @Override
    public void handlerRemoved(ChannelHandlerContext ctx) throws Exception {
        log.info("{}-----》客户端手动退出",  process.getAddress(ctx.channel()));
        //telnet 不会进入channelRead0方法，因此不会构建handshaker
        if (handShaker != null){
            CloseWebSocketFrame fram = new CloseWebSocketFrame();
            handShaker.close(ctx.channel(), fram.retain());
        }
        process.logout(ctx.channel());
    }

    @Override
    public void channelInactive(ChannelHandlerContext ctx) throws Exception {
        log.info("{}-----》客户端主动断链",  process.getAddress(ctx.channel()));
    }

    @Override
    public void exceptionCaught(ChannelHandlerContext ctx, Throwable cause) throws Exception {
//        super.exceptionCaught(ctx, cause);
        log.info("{}---客户端异常---->{}", process.getAddress(ctx.channel()), cause.getMessage());
        cause.printStackTrace();
//        process.logout(ctx.channel());
    }
}
