package cn.com.ty.chat.neety.support;

import cn.com.ty.chat.neety.protocol.IMP;
import cn.com.ty.chat.service.IChatSessionService;
import cn.com.ty.chat.service.ILiftCaseService;
import cn.com.ty.chat.service.impl.ChatSessionServiceImpl;
import cn.com.ty.chat.service.impl.LiftCaseServiceImpl;
import cn.com.ty.chat.neety.protocol.IMEncoder;
import cn.com.ty.chat.neety.protocol.IMMessage;
import cn.com.ty.chat.utils.ApplicationContextUtil;
import cn.com.ty.chat.utils.RedisUtils;
import com.alibaba.fastjson.JSONObject;
import io.netty.channel.Channel;
import io.netty.channel.group.ChannelGroup;
import io.netty.channel.group.DefaultChannelGroup;
import io.netty.handler.codec.http.websocketx.TextWebSocketFrame;
import io.netty.util.AttributeKey;
import io.netty.util.concurrent.GlobalEventExecutor;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.lang3.StringUtils;
import org.springframework.context.annotation.DependsOn;
import org.springframework.context.annotation.Lazy;
import org.springframework.stereotype.Component;

import java.util.*;
import java.util.concurrent.ConcurrentHashMap;
import java.util.function.Function;
import java.util.function.Predicate;

/**
 * websoket 服务支持类
 * 
 * @author sKing
 * @Date 2020年6月12日
 */
@Slf4j
public class WebSocketSupport {

    /** 用户名 */
    protected final AttributeKey<String> USERID = AttributeKey.valueOf("USERID");
    /** 用户名 */
    protected final AttributeKey<String> USERNAME = AttributeKey.valueOf("username");
    /** 头像 */
    protected final AttributeKey<String> HEAD_PIC = AttributeKey.valueOf("headPic");
    /** ip */
    protected final AttributeKey<String> IP_ADDR = AttributeKey.valueOf("ipAddr");

    protected final AttributeKey<JSONObject> ATTRS = AttributeKey.valueOf("attrs");
    /** 房间号 */
    protected final AttributeKey<String> ROOM = AttributeKey.valueOf("ROOM");
    /** 当前用户id */
    protected final AttributeKey<String> SENDID = AttributeKey.valueOf("SENDID");
    //用户与连接关系 key==userId
    protected static Map<String, List<Channel>> userChannel = new ConcurrentHashMap<>();

    /** 小房间用户关系对应 */
    protected static Map<String, Set<Channel>> roomUser = new HashMap<>();
    //房间与用户之间的关系，
    protected static Map<String, Set<HashMap<String, Channel>>> roomUserChannel = new ConcurrentHashMap<>();
    /** redis缓存 */
    protected static RedisUtils redis = ApplicationContextUtil.getBean(RedisUtils.class);


    protected static ILiftCaseService liftCaseService = ApplicationContextUtil.getBean(LiftCaseServiceImpl.class);
    /** 解析模式（本地测试为：false） */
    protected final static boolean testFlag = true;

    /** redis大房间key：ROOM */
    protected final String MAX_ROOM = "ROOM";

    /**
     * 获取用户昵称
     * 
     * @param client
     * @return
     */
    public String getNickName(Channel client) {
        return client.attr(USERNAME).get();
    }

    /**
     * 获取用户远程IP地址
     * 
     * @param client
     * @return
     */
    public String getAddress(Channel client) {
        return client.remoteAddress().toString().replaceFirst("/", "");
    }

    /**
     * 获取扩展属性
     * 
     * @param client
     * @return
     */
    protected JSONObject getAttrs(Channel client) {
        try {
            return client.attr(ATTRS).get();
        } catch (Exception e) {
            return null;
        }
    }

    /**
     * 获取扩展属性
     * 
     * @param client
     * @return
     */
    protected void setAttrs(Channel client, String key, Object value) {
        try {
            JSONObject json = client.attr(ATTRS).get();
            json.put(key, value);
            client.attr(ATTRS).set(json);
        } catch (Exception e) {
            JSONObject json = new JSONObject();
            json.put(key, value);
            client.attr(ATTRS).set(json);
        }
    }

    /**
     * 登出通知
     * 
     * @param client
     */
    public void logout(Channel client) {
        String userId = client.attr(USERID).get();
        if (StringUtils.isBlank(userId)){
            client.pipeline().close();
            return;
        }
        if (!userChannel.containsKey(userId)){
            client.pipeline().close();
            return;
        }

        List<Channel> channelList = userChannel.get(userId);
        if (!channelList.contains(client)){
            client.pipeline().close();
            return;
        }

        Iterator<Channel> iterator = channelList.iterator();
        while (iterator.hasNext()){
            Channel channel = iterator.next();
            if (channel == client){
                iterator.remove();
                client.pipeline().close();
                return;
            }
        }
    }

    /**
     * 系统当前时间
     * 
     * @return
     */
    protected long sysTime() {
        return System.currentTimeMillis();
    }

    protected static <T> Predicate<T> distinctByKey(Function<? super T, ?> keyExtractor) {
        Set<Object> seen = ConcurrentHashMap.newKeySet();
        return t -> seen.add(keyExtractor.apply(t));
    }

    /**
     * 向房间内在线用户发送其它人的在线状况支持
     * 
     * @param channel
     * @param loginList
     */
    protected void sendOnline(Channel channel, String loginList) {
        // 自定义IM协议解码
        IMMessage request = new IMMessage(IMP.SYSTEM.getName(), sysTime(), 0, loginList);
        String text = IMEncoder.toJsonString(request, testFlag);
        // 发送消息
        channel.writeAndFlush(new TextWebSocketFrame(text));
    }

    /**
     * 发送文本消息
     * 
     * @param channel
     * @param request
     */
    protected void sendMessage(Channel channel, IMMessage request) {
        // 自定义IM协议解码
        String text = IMEncoder.toJsonString(request, testFlag);
        log.info("给channel：{}发送消息",channel.id());
        // 发送消息
        channel.writeAndFlush(new TextWebSocketFrame(text));
    }

}
